// Author: Vladimir Savic
// E-mail: vladimirsvc@yahoo.com

#include "DataTypes.h"
#include "StateMachine.h"
#include "StateAbstract.h"
#include "StateMachineAbstract.h"
#include "TransitionMatrix.h"

// ***************************************************************************************************************************************************
// *  TRANSITION MATRIX STRUCTURE:                                                                                                                   *
// *    <-                                 NUMBER_OF_TM_COLUMNS                                   ->                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// *    |     DT     |      PS    |     TPS    |    Cmd1    |    Cmd2    |    Cmd3    |    Cmd4    |                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// * S1 |<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// * S2 |<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// * S3 |<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// * S4 |<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// * S5 |<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// * S6 |<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|<- xxMode ->|                                                 *
// *    --------------------------------------------------------------------------------------------                                                 *
// *  xxMode - depends on number of states// *                                                                                                       *
// *  TRANSITION MATRIX LEGEND:                                                                                                                      *
// *  PS   - Previous state                                                                                                                          *
// *  TPS  - Transition by PS  (state machine goes directly to next state only if PS condition is satisfied)                                         *
// *  DT   - Direct transition (when default method is executed, state machine goes directly to another state)                                       *
// *  CmdX - where X belongs to the set of N numbers, present external commands, also called signals/events                                          *
// *  Sx   - where x belongs to the set of N numbers, present state of the state machine                                                             *
// *                                                                                                                                                 *
// *  TRANSITION RULES:                                                                                                                              *
// *  -----------------                                                                                                                              *
// *  Transition types:                                                                                                                              *
// *  PS, TPS  - conditional transition, transition to another state depends on previous state.                                                      *
// *             If previous state is defined TPS field determines next state, all external commands are excluded                                    *
// *  DT       - direct transition, if direct transition is set all external commands are excluded including conditional transition                  *
// *  CmdX     - external command transition. External transitions are included only in case when DT is set to 0xFF as well as PS                    *
// *  Priorities are set at initialization time for each state, priorities are not fixed like for previous versions of the state machine             *
// *                                                                                                                                                 *
// *  TRANSITION MATRIX FILEDS DEFINITION:                                                                                                           *
// *  each transition matrix filed is determined by encoding type: x1 Mode, x2 Mode (for more details look at TransitionMatrix.h                     *
// *                                                                                                                                                 *
// *  ENCODING:                                                                                                                                      *
// *  each state/cmd must be encoded (identified), encoding depends on transition matrix encoding mode (for more details look at TransitionMatrix.h) *
// *                                                                                                                                                 *
// *  states encoding example:                                                                                                                       *
// *  S1 -> 0x00 ( first  transition matrix row, index 0 )                                                                                           *
// *  S2 -> 0x01 ( second transition matrix row, index 1 )                                                                                           *
// *  S3 -> 0x02 ( third  transition matrix row, index 2 )                                                                                           *
// *  S4 -> 0x03 ( fourth transition matrix row, index 3 )                                                                                           *
// *                                                                                                                                                 *
// *  commands encoding (0x00, 0x01 and 0x02 cannot be used for command encoding - those codes are reserved for DT, PS and TPS respectively):        *
// *  encoding for external commands starts from 0x03 ... These encodings determines transition matrix index (look at TransitionMatrix.h)            *
// *  ---------------------------------------------------------------------------------------------------------------------------------------        *
// *  DT   -> always has 0x00 code, transition matrix column index 0                                                                                 *
// *  PS   -> always has 0x01 code, transition matrix column index 1                                                                                 *
// *  TPS  -> always has 0x02 code, transition matrix column index 2                                                                                 *
// *  ---------------------------------------------------------------------------------------------------------------------------------------        *
// *  Cmd1 -> 0x03 external signal                                                                                                                   *
// *  Cmd2 -> 0x04 external signal                                                                                                                   *
// *  Cmd3 -> 0x05 external signal                                                                                                                   *
// *  Cmd4 -> 0x06 external signal                                                                                                                   *
// *                                                                                                                                                 *
// *  SERIAL COMMUNICATION BUFFER STREAM (for this example):                                                                                         *
// *  Number of bytes required for TM depends on state encoding mode                                                                                 *
// *  1  byte required for start state identificator - S0                                                                                            *
// *  -----------------------------                                                                                                                  *
// *  |    S0    |     TM         |                                                                                                                  *
// *  -----------------------------                                                                                                                  *
// *  |<- BYTE ->|<- xMode_BYTES ->                                                                                                                  *
// *                                                                                                                                                 *
// ***************************************************************************************************************************************************

// Global constants:
const BYTE BUFFER_SIZE = 17;                                          // serial communication buffer size, x4Mode is used

// *******************************
// * Global variable declaration *
// *******************************

BYTE idata buffer[BUFFER_SIZE] ;
BYTE bufferIndex               ;
BYTE command                   ;

StateMachineAbstract    stateMachine     ;
TransitionMatrix        transitionMatrix ;

STEPPER_MOTOR_DATA_TYPE StepperMotor     ;

// ********************
// * Main entry point *
// ********************
void main(){
    UART1_Init( 4800 ) ;                                              // RS 232 communication Initialization
    Delay_ms(500)      ;
    
    // Stepper Motor Settings:
    StepperMotor.ControlSignals   = 0x00 ;
    StepperMotor.AngleCounter     =    0 ;
    StepperMotor.LeftAnglePreset  =  -30 ;
    StepperMotor.RightAnglePreset =   30 ;
    StepperMotor.TimeSlice        =  500 ;

    bufferIndex = 0x00 ;
    command     = 0x00 ;
    
                                                                      // stepper motor transition matrix initialization
    transitionMatrix.startState     = &buffer[0];                     // S0  pointer
    transitionMatrix.buffer         = &buffer[1];                     // TM  pointer
    transitionMatrix.Initialization = &PriorityMethodsInitialization; // priority matrix Initialization
    transitionmatrix.initialization( &transitionMatrix );             // execute initialization methods: set priority methods

    stateMachine.Initialization = &Initialization;                    // state machine Initialization
    stateMachine.Initialization( &stateMachine ) ;
    stateMachine.StateInitialization( )          ;
    stateMachine.TransitionMatrixInitialization(&transitionMatrix);   // initialize state machine with transition matrix
    
    while(1){
        if ( UART1_Data_Ready() == 1 ){
            buffer[bufferIndex++] = UART1_Read();
        }
        else if (bufferIndex == BUFFER_SIZE){
            bufferIndex = 0;
            stateMachine.StartProcessing( );
        }
        else if(stateMachine.IsStarted == 0xFF){
            stateMachine.ProcessCommand( );
        }
        
        P2 = StepperMotor.ControlSignals;
        command = P3 >> 5;
    }
}