// Author: Vladimir Savic
// E-mail: vladimirsvc@yahoo.com

#include "TransitionMatrix.h"

static TransitionMatrix* ptrTransitionMatrix ;
static BYTE currentStateIndex ;
static BYTE pStateIndex    ;
static BYTE currentStateFlag  ;
static BYTE externalCommand   ;

// This is the right way, but it takes a lot of memory
/*BYTE ( *FirstMethodPriority  ) ( );
  BYTE ( *SecondMethodPriority ) ( );
  BYTE ( *ThirdMethodPriority  ) ( );*/

// *****************************************************************************
// * Transition matrix priority methods initialization                         *
// * at this point, set priority levels for DT, PS and CMD state transitions   *
// *****************************************************************************
void PriorityMethodsInitialization ( void* ptr ){
    ptrTransitionMatrix = (TransitionMatrix*)ptr;

    // method initialization:
    ptrTransitionMatrix->GetNextState = &GetNextState;
   
    ptrTransitionMatrix->xMode      = 0x0F; // x2Mode
    ptrTransitionMatrix->xModeBytes = 0x04; // 4bytes for one state

   // NOTE: !!!
    // Code below present generic way, to set priority execution between different types of transitions
    // This is the right way to implement priorities but it takes a lot of memory
    // Instead, priority of execution is set manually, look at GetNextState method
    
    // First method to execute : DT
    // Second method to execute: PS
    // Third method to execute : CMD
    /*FirstMethodPriority  = &X2ModeIsDTTransitionAvailable ;
      SecondMethodPriority = &X2ModeIsPSTransitionAvailable ;
      ThirdMethodPriority  = &X2ModeIsCmdTransitionAvailable;*/
    
    // it could be also set like this:
    // First method to execute : CMD
    // Second method to execute: PS
    // Third method to execute : DT
    // FirstMethodPriority  = &X2ModeIsCmdTransitionAvailable ;
    // SecondMethodPriority = &X2ModeIsPSTransitionAvailable  ;
    // ThirdMethodPriority  = &X2ModeIsDTTransitionAvailable  ;
}

// *****************************************************************************
// * Get next state by current state info and external command                 *
// * following priority levels. Priority levels are set at initialization time *
// *****************************************************************************
static BYTE GetNextState ( BYTE stateIndex, BYTE prevStateIndex, BYTE stateFlag, BYTE cmd ){
    BYTE newState = 0xFF;

    // current state index, current state flag and external commands are used by priority methods
    currentStateIndex = stateIndex     ;
    pStateIndex       = prevStateIndex ;
    currentStateFlag  = stateFlag      ;
    externalCommand   = cmd            ;

    newState = CalcNextState( currentStateFlag & 0x03 );
    if( newState == ptrTransitionMatrix->xMode ) newState = CalcNextState( (currentStateFlag & 0x0C) >> 2 );
    if( newState == ptrTransitionMatrix->xMode ) newState = CalcNextState( (currentStateFlag & 0x30) >> 4 );

    return newState;
}

static BYTE CalcNextState( BYTE prioritCode ){
    BYTE newState = 0xFF;
    
    switch(prioritCode){
        case 0x00:
            newState = X2ModeIsDTTransitionAvailable ( );
        break;
        case 0x01:
            newState = X2ModeIsPSTransitionAvailable ( );
        break;
        case 0x02:
            newState = X2ModeIsCmdTransitionAvailable( );
        break;
        default: ;
    }
    
    return newState;
}

// TODO: x1 mode
/*static BYTE X1ModeIsDTTransitionAvailable( ){
}
static BYTE X1ModeIsPSTransitionAvailable( ){
}
static BYTE X1ModeIsCmdTransitionAvailable( ){
}*/

// ********************************************************
// * x2 encoding mode - two state codes in one BYTE       *
// * each state is encoded with 4 bits                    *
// * maximum number of states: 16 - 1                     *
// * -------------------------                            *
// * |<-        BYTE       ->|                            *
// * -------------------------                            *
// * |<- 4bits ->|<- 4bits ->|                            *
// * -------------------------                            *
// ********************************************************

// -----------------------------------------------------------------------------------------
// |      x2 Index 0     |      x2 Index 1     |      x2 Index 2     |      x2 Index 3     |     x2 Index
// -----------------------------------------------------------------------------------------
// |     0    |      1   |     0    |      1   |     0    |     1    |     0    |     1    |     Position {0,1}
// -----------------------------------------------------------------------------------------
// |<-  4bits |  4bits ->|<-  4bits |  4bits ->|<-  4bits |  4bits ->|<-  4bits |  4bits ->|     4 BYTES - x2 ENCODING MODE -> for 4 states: total bytes required = 4*4 = 16 bytes
// -----------------------------------------------------------------------------------------
// |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |  NOTHING |     7 BYTES - x1 ENCODING MODE -> for 4 states: total bytes required = 4*7 = 28 bytes
// -----------------------------------------------------------------------------------------
// | Index: 0 | Index: 1 | Index: 2 | Index: 3 | Index: 4 | Index: 5 | Index: 6 |  NOTHING |     matixPosition (look at TransitionMatrix.h)
// -----------------------------------------------------------------------------------------

// x2 Index mapping: indexX / 2  = x2IndexX
// position inside BYTE: indexX mod 2 = {0,1}

// example1: indexX = 5 (Cmd3 external command)
//           it must be mapped to: x2Index2, position 1
//           5/2 = 2; 5 mod 2 = 1; -> x2Index2, position 1

// example2: indexX = 6 (Cmd4 external command)
//           it must be mapped to: x2Index3, position 0
//           6/2 = 3; 6 mod 2 = 0; -> x2Index3, position 0

// ***************************************
// * Is DT (direct transition) available *
// ***************************************
static BYTE X2ModeIsDTTransitionAvailable( ){
    BYTE result    = ptrTransitionMatrix->xMode;
    BYTE DTAllowed = 0x00;

    DTAllowed = currentStateFlag & 0x80;

    if( DTAllowed != 0x00 ){
        // we know that DT filed has v2Index 0 and position 0
        // no need for complete set of calculations like for external commands (cmd)
        result = *(ptrTransitionMatrix->buffer + (ptrTransitionMatrix->xModeBytes * currentStateIndex)) & 0xF0;
        result >>= 4;
    }

    return result;
}

// ********************************************
// * Is PS (conditional transition) available *
// ********************************************
static BYTE X2ModeIsPSTransitionAvailable( ){
    BYTE result     = ptrTransitionMatrix->xMode;
    BYTE TPSAllowed = 0x00;

    TPSAllowed = currentStateFlag & 0x40;

    if( TPSAllowed != 0x00 ){
        // we know that PS filed has v2Index 0 and position 1
        // no need for complete set of calculations like for external commands (cmd)
        result = *(ptrTransitionMatrix->buffer + (ptrTransitionMatrix->xModeBytes * currentStateIndex)) & 0x0F;
        if(result != 0x0F && result == pStateIndex){
            result = *((ptrTransitionMatrix->buffer + ((ptrTransitionMatrix->xModeBytes * currentStateIndex) + 1))/*TPS x2Index is 1*/) & 0xF0; // extract TPS
            result >>= 4;
        }
        else{
           result = ptrTransitionMatrix->xMode;
        }
    }

    return result;
}

// **************************************************
// * Is Cmd (external command transition) available *
// **************************************************
static BYTE X2ModeIsCmdTransitionAvailable( ){
    BYTE result = ptrTransitionMatrix->xMode;

    if( externalCommand > 0x02 && externalCommand < (NUMBER_OF_EXT_CMD + 0x03)){
        if((externalCommand % 2) != 0x00)
            result = *((ptrTransitionMatrix->buffer + (ptrTransitionMatrix->xModeBytes * currentStateIndex)) + (externalCommand / 2)) & 0x0F /*set to position 1*/;
        else
            result = *((ptrTransitionMatrix->buffer + (ptrTransitionMatrix->xModeBytes * currentStateIndex)) + (externalCommand / 2)) & 0xF0 /*set to position 0*/;

        if((externalCommand % 2) == 0x00) result >>= 4;
    }

    return result;
}